#include "asf_vector.h"
#include "shapefil.h"
#include "asf_nan.h"
#include <assert.h>
#include <errno.h>
#include <ctype.h>

typedef struct {
  char filename[1024];
  char date_photography[64];
  char location[128];
  char mission_line[128];
  char roll[64];
  int frame;
  double center_lat;
  double center_lon;
  double near_start_lat;
  double near_start_lon;
  double far_start_lat;
  double far_start_lon;
  double near_end_lat;
  double near_end_lon;
  double far_end_lat;
  double far_end_lon;
  char photography[128];
  char camera[128];
  char focal_length[128];
  char sensor[128];
  char altitude[128];
  char field_of_view[128];
  char resolution[128];
  char overlap[128];
  char scale[128];
  char comments[128];
  char metadata[1024];
  char *thumbnail;
  char *description;
  char *download;
  double size;
  char *download2;
  double size2;
} hap_t; // high altitude photography

void hap_init(hap_t *hap)
{
  strcpy(hap->filename, MAGIC_UNSET_STRING);
  strcpy(hap->date_photography, MAGIC_UNSET_STRING);
  strcpy(hap->location, MAGIC_UNSET_STRING);
  strcpy(hap->mission_line, MAGIC_UNSET_STRING);
  strcpy(hap->roll, MAGIC_UNSET_STRING);
  hap->frame = MAGIC_UNSET_INT;
  hap->center_lat = MAGIC_UNSET_DOUBLE;
  hap->center_lon = MAGIC_UNSET_DOUBLE;
  hap->near_start_lat = MAGIC_UNSET_DOUBLE;
  hap->near_start_lon = MAGIC_UNSET_DOUBLE;
  hap->far_start_lat = MAGIC_UNSET_DOUBLE;
  hap->far_start_lon = MAGIC_UNSET_DOUBLE;
  hap->near_end_lat = MAGIC_UNSET_DOUBLE;
  hap->near_end_lon = MAGIC_UNSET_DOUBLE;
  hap->far_end_lat = MAGIC_UNSET_DOUBLE;
  hap->far_end_lon = MAGIC_UNSET_DOUBLE;
  strcpy(hap->photography, MAGIC_UNSET_STRING);
  strcpy(hap->camera, MAGIC_UNSET_STRING);
  strcpy(hap->focal_length, MAGIC_UNSET_STRING);
  strcpy(hap->sensor, MAGIC_UNSET_STRING);
  strcpy(hap->altitude, MAGIC_UNSET_STRING);
  strcpy(hap->field_of_view, MAGIC_UNSET_STRING);
  strcpy(hap->resolution, MAGIC_UNSET_STRING);
  strcpy(hap->overlap, MAGIC_UNSET_STRING);
  strcpy(hap->scale, MAGIC_UNSET_STRING);
  strcpy(hap->comments, MAGIC_UNSET_STRING);
  strcpy(hap->metadata, MAGIC_UNSET_STRING);
  hap->thumbnail = NULL;
  hap->description = NULL;
  hap->download = NULL;
  hap->size = MAGIC_UNSET_DOUBLE;
  hap->download2 = NULL;
  hap->size2 = MAGIC_UNSET_DOUBLE;
}

void hap_free(hap_t *hap)
{
  if (hap->thumbnail) {
    FREE(hap->thumbnail);
    hap->thumbnail = NULL;
  }
  if (hap->description) {
    FREE(hap->description);
    hap->description = NULL;
  }
  if (hap->download) {
    FREE(hap->download);
    hap->download = NULL;
  }
  if (hap->download2) {
    FREE(hap->download2);
    hap->download2 = NULL;
  }
}

static void add_to_kml(FILE *fp, hap_t *hap, dbf_header_t *dbf, int nCols)
{
  int ii;
  char begin[10], end[10], line[1024];
  FILE *fpIn;

  // Print out according to configuration
  fprintf(fp, "<Placemark>\n");
  fprintf(fp, "  <description><![CDATA[\n");
  if (hap->thumbnail && hap->description)
    fprintf(fp, "<table width=\"600\">\n");
  else
    fprintf(fp, "<table width=\"350\">\n");
  if (hap->thumbnail || hap->description)
    fprintf(fp, "<tr>\n");
  if (hap->thumbnail)
    fprintf(fp, "<td valign=\"top\"><img src=\"%s\" width=\"256\" "
	    "height=\"256\"><br><br>\n", hap->thumbnail);
  if (hap->download) {
    fprintf(fp, "<strong>Download data:</strong><br>\n");
    fprintf(fp, "<strong><a href=\"%s\">TIFF format (%.1f MB)</a></strong>"
	    "<br>", hap->download, hap->size);
    if (hap->download2)
      fprintf(fp, "<strong><a href=\"%s\">JPEG format (%.1f MB)</a></strong>"
	      "<br>", hap->download2, hap->size2);
  }
  if (hap->thumbnail)
    fprintf(fp, "</td><td with=\"10\">&nbsp;</td>\n");
  if (hap->description) {
    fpIn = FOPEN(hap->description, "r");
    fprintf(fp, "<td>");
    while (fgets(line, 1022, fpIn) != NULL)
      fprintf(fp, "%s", line);
    fprintf(fp, "<br><br>\n");
  }
  else if (hap->thumbnail || hap->description)
    fprintf(fp, "</td></tr>\n");
  if (hap->thumbnail && hap->description)
    fprintf(fp, "<tr><td colspan=\"3\">\n");
  else
    fprintf(fp, "<tr><td>\n");
  fprintf(fp, "<!-- Format: HAP (generated by %s)-->\n", version_string("convert2vector"));
  fprintf(fp, "<strong>Corner coordinates:</strong><br>\n");
  fprintf(fp, "<table cellpadding=\"0\" cellspacing=\"0\">\n");
  for (ii=0; ii<nCols; ii++) {
    if (dbf[ii].visible == 0) {
      strcpy(begin, "<!--");
      strcpy(end, "-->\n");
    }
    else {
      strcpy(begin, "");
      strcpy(end, "\n");
    }
    if (strcmp(dbf[ii].header, "Upper_Left_Lat") == 0)
      fprintf(fp, "<tr><td width=\"200\"><strong>Upper left lat</strong>: %s"
	      "</td>", lf(hap->near_start_lat));
    else if (strcmp(dbf[ii].header, "Upper_Left_Lon") == 0)
      fprintf(fp, "<td><strong>Upper left lon</strong>: %s</td></tr>\n",
	      lf(hap->near_start_lon));
    else if (strcmp(dbf[ii].header, "Upper_Right_Lat") == 0)
      fprintf(fp, "<tr><td width=\"200\"><strong>Upper right lat</strong>: %s"
	      "</td>", lf(hap->far_start_lat));
    else if (strcmp(dbf[ii].header, "Upper_Right_Lon") == 0)
      fprintf(fp, "<td><strong>Upper right lon</strong>: %s</td></tr>\n", 
	      lf(hap->far_start_lon));
    else if (strcmp(dbf[ii].header, "Lower_Left_Lat") == 0)
      fprintf(fp, "<tr><td width=\"200\"><strong>Lower left lat</strong>: %s"
	      "</td>", lf(hap->near_end_lat));
    else if (strcmp(dbf[ii].header, "Lower_Left_Lon") == 0)
      fprintf(fp, "<td><strong>Lower left lon</strong>: %s</td></tr>\n", 
	      lf(hap->near_end_lon));
    else if (strcmp(dbf[ii].header, "Lower_Right_Lat") == 0)
      fprintf(fp, "<tr><td width=\"200\"><strong>Lower right lat</strong>: %s"
	      "</td>", lf(hap->far_end_lat));
    else if (strcmp(dbf[ii].header, "Lower_Right_Lon") == 0)
      fprintf(fp, "<td><strong>Lower right lon</strong>: %s</td></tr>\n", 
	      lf(hap->far_end_lon));
    /*
    if (strcmp(dbf[ii].header, "Date_Photography") == 0)
      fprintf(fp, "%s<strong>Date of Photography</strong>: %s <br>%s", 
	      begin, hap->date_photography, end);
    else if (strcmp(dbf[ii].header, "Location") == 0)
      fprintf(fp, "%s<strong>Location</strong>: %s <br>%s", 
	      begin, hap->location, end);
    else if (strcmp(dbf[ii].header, "Mission_Line") == 0)
      fprintf(fp, "%s<strong>Mission/Line</strong>: %s <br>%s",
	      begin, hap->mission_line, end);
    else if (strcmp(dbf[ii].header, "Roll") == 0)
      fprintf(fp, "%s<strong>Roll</strong>: %s <br>%s",
	      begin, hap->roll, end);
    else if (strcmp(dbf[ii].header, "Frame") == 0)
      fprintf(fp, "%s<strong>Frame</strong>: %d <br>%s",
	      begin, hap->frame, end);
    else if (strcmp(dbf[ii].header, "Upper_Left_Lat") == 0)
      fprintf(fp, "%s<strong>Upper left lat</strong>: %s <br>%s", 
	      begin, lf(hap->near_start_lat), end);
    else if (strcmp(dbf[ii].header, "Upper_Left_Lon") == 0)
      fprintf(fp, "%s<strong>Upper left lon</strong>: %s <br>%s", 
	      begin, lf(hap->near_start_lon), end);
    else if (strcmp(dbf[ii].header, "Upper_Right_Lat") == 0)
      fprintf(fp, "%s<strong>Upper right lat</strong>: %s <br>%s", 
	      begin, lf(hap->far_start_lat), end);
    else if (strcmp(dbf[ii].header, "Upper_Right_Lon") == 0)
      fprintf(fp, "%s<strong>Upper right lon</strong>: %s <br>%s", 
	      begin, lf(hap->far_start_lon), end);
    else if (strcmp(dbf[ii].header, "Lower_Left_Lat") == 0)
      fprintf(fp, "%s<strong>Lower left lat</strong>: %s <br>%s", 
	      begin, lf(hap->near_end_lat), end);
    else if (strcmp(dbf[ii].header, "Lower_Left_Lon") == 0)
      fprintf(fp, "%s<strong>Lower left lon</strong>: %s <br>%s", 
	      begin, lf(hap->near_end_lon), end);
    else if (strcmp(dbf[ii].header, "Lower_Right_Lat") == 0)
      fprintf(fp, "%s<strong>Lower right lat</strong>: %s <br>%s", 
	      begin, lf(hap->far_end_lat), end);
    else if (strcmp(dbf[ii].header, "Lower_Right_Lon") == 0)
      fprintf(fp, "%s<strong>Lower right lon</strong>: %s <br>%s", 
	      begin, lf(hap->far_end_lon), end);
    else if (strcmp(dbf[ii].header, "Photography") == 0)
      fprintf(fp, "%s<strong>Photography</strong>: %s <br>%s",
	      begin, hap->photography, end);
    else if (strcmp(dbf[ii].header, "Camera") == 0)
      fprintf(fp, "%s<strong>Camera</strong>: %s <br>%s",
	      begin, hap->camera, end);
    else if (strcmp(dbf[ii].header, "Focal_Length") == 0)
      fprintf(fp, "%s<strong>Camera focal length</strong>: %s <br>%s",
	      begin, hap->focal_length, end);
    else if (strcmp(dbf[ii].header, "Sensor_Coverage") == 0)
      fprintf(fp, "%s<strong>Sensor coverage</strong>: %s <br>%s",
	      begin, hap->sensor, end);
    else if (strcmp(dbf[ii].header, "Altitude") == 0)
      fprintf(fp, "%s<strong>Altitude of acquisition</strong>: %s <br>%s",
	      begin, hap->altitude, end);
    else if (strcmp(dbf[ii].header, "Field_Of_View") == 0)
      fprintf(fp, "%s<strong>Horizontal field of view</strong>: %s <br>%s",
	      begin, hap->field_of_view, end);
    else if (strcmp(dbf[ii].header, "Resolution") == 0)
      fprintf(fp, "%s<strong>Resolution</strong>: %s <br>%s",
	      begin, hap->resolution, end);
    else if (strcmp(dbf[ii].header, "Overlap") == 0)
      fprintf(fp, "%s<strong>Overlap</strong>: %s <br>%s",
	      begin, hap->overlap, end);
    else if (strcmp(dbf[ii].header, "Scale") == 0)
      fprintf(fp, "%s<strong>Scale</strong>: %s <br>%s",
	      begin, hap->scale, end);
    else if (strcmp(dbf[ii].header, "Comments") == 0)
      fprintf(fp, "%s<strong>Comments</strong>: %s <br>%s",
	      begin, hap->comments, end);
    */
  }
  fprintf(fp, "</table><br>\n");
  fprintf(fp, "<strong>Download <a href=\"%s\">complete metadata</a></strong>"
	  "<br>\n", hap->metadata);
  fprintf(fp, "</td></tr></table>\n");
  fprintf(fp, "  ]]></description>\n");
  fprintf(fp, "  <name>%s</name>\n", hap->filename);
  fprintf(fp, "  <LookAt>\n");
  fprintf(fp, "    <longitude>%.10f</longitude>\n", hap->center_lon);
  fprintf(fp, "    <latitude>%.10f</latitude>\n", hap->center_lat);
  fprintf(fp, "    <range>400000</range>\n");
  fprintf(fp, "  </LookAt>\n");
  fprintf(fp, "  <visibility>1</visibility>\n");
  fprintf(fp, "  <open>1</open>\n");

  write_kml_style_keys(fp);

  fprintf(fp, "  <Polygon>\n");
  fprintf(fp, "    <extrude>1</extrude>\n");
  fprintf(fp, "    <altitudeMode>%s</altitudeMode>\n", altitude_mode());
  fprintf(fp, "    <outerBoundaryIs>\n");
  fprintf(fp, "     <LinearRing>\n");
  fprintf(fp, "      <coordinates>\n");
  fprintf(fp, "       %.12f,%.12f,7000\n", 
	  hap->near_start_lon, hap->near_start_lat);
  fprintf(fp, "       %.12f,%.12f,7000\n", 
	  hap->far_start_lon, hap->far_start_lat);
  fprintf(fp, "       %.12f,%.12f,7000\n", 
	  hap->far_end_lon, hap->far_end_lat);
  fprintf(fp, "       %.12f,%.12f,7000\n", 
	  hap->near_end_lon, hap->near_end_lat);
  fprintf(fp, "       %.12f,%.12f,7000\n", 
	  hap->near_start_lon, hap->near_start_lat);
  fprintf(fp, "      </coordinates>\n");
  fprintf(fp, "     </LinearRing>\n");
  fprintf(fp, "    </outerBoundaryIs>\n");
  fprintf(fp, "  </Polygon>\n");
  fprintf(fp, "</Placemark>\n");
}

static int read_hap_line(char *header, int n,char *line, hap_t *hap)
{
  int ii, ok;
  char *test = (char *) MALLOC(sizeof(char)*255);
  for (ii=0; ii<n; ii++) {
    test = get_column(header, ii);
    if (strcmp(test, "Filename") == 0)
      strcpy(hap->filename, get_str(line, ii));
    else if (strcmp(test, "Date_Photography") == 0)
      strcpy(hap->date_photography, get_str(line, ii));
    else if (strcmp(test, "Location") == 0)
      strcpy(hap->location, get_str(line, ii));
    else if (strcmp(test, "Mission_Line") == 0)
      strcpy(hap->mission_line, get_str(line, ii));
    else if (strcmp(test, "Roll") == 0)
      strcpy(hap->roll, get_str(line, ii));
    else if (strcmp(test, "Frame") == 0)
      hap->frame = get_int(line, ii);
    else if (strcmp(test, "Center_Lat") == 0)
      hap->center_lat = get_double(line, ii);
    else if (strcmp(test, "Center_Lon") == 0)
      hap->center_lon = get_double(line, ii);
    else if (strcmp(test, "Upper_Left_Lat") == 0)
      hap->near_start_lat = get_req_double(line, ii, &ok);
    else if (strcmp(test, "Upper_Left_Lon") == 0)
      hap->near_start_lon = get_req_double(line, ii, &ok);
    else if (strcmp(test, "Upper_Right_Lat") == 0) 
      hap->far_start_lat = get_req_double(line, ii, &ok);
    else if (strcmp(test, "Upper_Right_Lon") == 0)
      hap->far_start_lon = get_req_double(line, ii, &ok);
    else if (strcmp(test, "Lower_Left_Lat") == 0)
      hap->near_end_lat = get_req_double(line, ii, &ok);
    else if (strcmp(test, "Lower_Left_Lon") == 0)
      hap->near_end_lon = get_req_double(line, ii, &ok);
    else if (strcmp(test, "Lower_Right_Lat") == 0)
      hap->far_end_lat = get_req_double(line, ii, &ok);
    else if (strcmp(test, "Lower_Right_Lon") == 0)
      hap->far_end_lon = get_req_double(line, ii, &ok);
    else if (strcmp(test, "Photography") == 0)
      strcpy(hap->photography, get_str(line, ii));
    else if (strcmp(test, "Camera") == 0)
      strcpy(hap->camera, get_str(line, ii));
    else if (strcmp(test, "Focal_Length") == 0)
      strcpy(hap->focal_length, get_str(line, ii));
    else if (strcmp(test, "Sensor_Coverage") == 0)
      strcpy(hap->sensor, get_str(line, ii));
    else if (strcmp(test, "Altitude") == 0)
      strcpy(hap->altitude, get_str(line, ii));
    else if (strcmp(test, "Field_Of_View") == 0)
      strcpy(hap->field_of_view, get_str(line, ii));
    else if (strcmp(test, "Resolution") == 0)
      strcpy(hap->resolution, get_str(line, ii));
    else if (strcmp(test, "Overlap") == 0)
      strcpy(hap->overlap, get_str(line, ii));
    else if (strcmp(test, "Scale") == 0)
      strcpy(hap->scale, get_str(line, ii));
    else if (strcmp(test, "Comments") == 0)
      strcpy(hap->comments, get_str(line, ii));
    else if (strcmp(test, "Metadata") == 0)
      strcpy(hap->metadata, get_str(line, ii));
    else if (strcmp(test, "Thumbnail") == 0) {
      hap->thumbnail = (char *) MALLOC(sizeof(char)*1024);
      strcpy(hap->thumbnail, get_str(line, ii));
    }
    else if (strcmp(test, "Description") == 0) {
      hap->description = (char *) MALLOC(sizeof(char)*1024);
      strcpy(hap->description, get_str(line, ii));
    }
    else if (strcmp(test, "Download1") == 0) {
      hap->download = (char *) MALLOC(sizeof(char)*1024);
      strcpy(hap->download, get_str(line, ii));
    }
    else if (strcmp(test, "Size1") == 0) {
      hap->size = get_double(line, ii);
    }
    else if (strcmp(test, "Download2") == 0) {
      hap->download2 = (char *) MALLOC(sizeof(char)*1024);
      strcpy(hap->download2, get_str(line, ii));
    }
    else if (strcmp(test, "Size2") == 0) {
      hap->size2 = get_double(line, ii);
    }
  }
  FREE(test);

  return ok;
}

// Check location information
static int check_hap_location(FILE *ifp, char **header_line, int *n)
{
  dbf_header_t *dbf;
  int ii, nCols;
  char *header = (char *) MALLOC(sizeof(char)*1024);
  fgets(header, 1024, ifp);
  strip_end_whitesp_inplace(header);
  int nColumns = get_number_columns(header);
  
  // Read configuration file
  read_header_config("HAP", &dbf, &nCols);
  
  // ensure we have the columns we need
  int name_col = find_str(header, "Filename");
  int near_start_lat_col = find_str(header, "Upper_Left_Lat");
  int near_start_lon_col = find_str(header, "Upper_Left_Lon");
  int far_start_lat_col = find_str(header, "Upper_Right_Lat");
  int far_start_lon_col = find_str(header, "Upper_Right_Lon");
  int near_end_lat_col = find_str(header, "Lower_Left_Lat");
  int near_end_lon_col = find_str(header, "Lower_Left_Lon");
  int far_end_lat_col = find_str(header, "Lower_Right_Lat");
  int far_end_lon_col = find_str(header, "Lower_Right_Lon");
  
  // Check whether all visible columns are actually available in the file
  for (ii=0; ii<nCols; ii++) {
    if (find_str(header, dbf[ii].header) < 0)
      dbf[ii].visible = FALSE;
  }
  
  int all_ok=TRUE;
  if (name_col < 0) {
    printf("Missing: Filename\n");
    all_ok=FALSE;
  }
  if (near_start_lat_col < 0) {
    printf("Missing: Upper_Left_Lat\n");
    all_ok=FALSE;
  }
  if (near_start_lon_col < 0) {
    printf("Missing: Upper_Left_Lon\n");
    all_ok=FALSE;
  }
  if (far_start_lat_col < 0) {
    printf("Missing: End_Start_Lat\n");
    all_ok=FALSE;
  }
  if (far_start_lon_col < 0) {
    printf("Missing: End_Start_Lon\n");
    all_ok=FALSE;
  }
  if (near_end_lat_col < 0) {
    printf("Missing: Lower_Left_Lat\n");
    all_ok=FALSE;
  }
  if (near_end_lon_col < 0) {
    printf("Missing: Lower_Left_Lon\n");
    all_ok=FALSE;
  }
  if (far_end_lat_col < 0) {
    printf("Missing: Lower_Right_Lat\n");
    all_ok=FALSE;
  }
  if (far_end_lon_col < 0) {
    printf("Missing: Lower_Right_Lon\n");
    all_ok=FALSE;
  }
  if (!all_ok) {
    printf("Required data columns missing, cannot process this file.\n");
    return 0;
  }
  *header_line = header;
  *n = nColumns;

  return 1;
}

// Convert hap to kml file
int hap2kml(char *in_file, char *out_file, int listFlag)
{
  hap_t hap;
  dbf_header_t *dbf;
  char *header;
  int nCols, nColumns;
  char line[1024];
  
  // Read configuration file
  read_header_config("HAP", &dbf, &nCols);

  FILE *ifp = FOPEN(in_file, "r");
  assert(ifp);
  check_hap_location(ifp, &header, &nColumns);

  FILE *ofp = FOPEN(out_file, "w");
  if (!ofp) {
    printf("Failed to open output file %s: %s\n", out_file, strerror(errno));
    return 0;
  }
  
  kml_header(ofp);
  
  if (listFlag) {
    while (fgets(line, 1022, ifp) != NULL) {
      strip_end_whitesp_inplace(line);
      
      // now get the individual column values
      hap_init(&hap);
      if (read_hap_line(header, nColumns, line, &hap))
	add_to_kml(ofp, &hap, dbf, nCols);
      hap_free(&hap);
    }
  }
  else {
    fgets(line, 1022, ifp);
    strip_end_whitesp_inplace(line);
    hap_init(&hap);
    if (read_hap_line(header, nColumns, line, &hap))
      add_to_kml(ofp, &hap, dbf, nCols);
    hap_free(&hap);
  }
  
  kml_footer(ofp);
  
  fclose(ifp);
  fclose(ofp);
  
  return 1;
}

void shape_hap_init(char *inFile, char *header)
{
  char *dbaseFile;
  DBFHandle dbase;
  SHPHandle shape;
  dbf_header_t *dbf;
  int ii, nCols, length=50;

  // Read configuration file
  read_header_config("HAP", &dbf, &nCols);

  // Open database for initialization
  dbaseFile = (char *) MALLOC(sizeof(char)*(strlen(inFile)+5));
  sprintf(dbaseFile, "%s.dbf", inFile);
  dbase = DBFCreate(dbaseFile);
  if (!dbase)
    asfPrintError("Could not create database file '%s'\n", dbaseFile);

  // Add fields to database
  for (ii=0; ii<nCols; ii++) {
    if (strcmp(dbf[ii].header, "Filename") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "FILENAME", FTString, length, 0) == -1)
        asfPrintError("Could not add FILENAME field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Date_Photography") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "DATE_PHOTO", FTString, length, 0) == -1)
        asfPrintError("Could not add DATE_PHOTO field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Location") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "LOCATION", FTString, length, 0) == -1)
        asfPrintError("Could not add LOCATION field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Mission_Line") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "MISSION", FTString, length, 0) == -1)
        asfPrintError("Could not add MISSION field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Roll") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "ROLL", FTString, length, 0) == -1)
        asfPrintError("Could not add ROLL field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Frame") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "FRAME", FTInteger, 16, 0) == -1)
        asfPrintError("Could not add FRAME field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Center_Lat") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_CLAT", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_CLAT field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Center_Lon") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_CLON", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_CLON field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Upper_Left_Lat") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_LULAT", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_LULAT field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Upper_Left_Lon") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_LULON", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_LULON field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Upper_Right_Lat") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_RULAT", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_RULAT field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Upper_Right_Lon") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_RULON", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_RULON field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Lower_Left_Lat") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_LDLAT", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_LDLAT field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Lower_Left_Lon") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_LDLON", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_LDLON field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Lower_Right_Lat") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_RDLAT", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_RDLAT field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Lower_Right_Lon") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCN_RDLON", FTDouble, 16, 7) == -1)
        asfPrintError("Could not add SCN_RDLON field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Photography") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "PHOTOGRAPHY", FTString, length, 0) == -1)
        asfPrintError("Could not add PHOTOGRAPHY field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Camera") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "CAMERA", FTString, length, 0) == -1)
        asfPrintError("Could not add CAMERA field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Focal_Length") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "FOCAL_LENGTH", FTString, length, 0) == -1)
        asfPrintError("Could not add FOCAL_LENGTH field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Sensor_Coverage") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SENSOR", FTString, length, 0) == -1)
        asfPrintError("Could not add SENSOR field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Altitude") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "ALTITUDE", FTString, length, 0) == -1)
        asfPrintError("Could not add ALTITUE field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Field_Of_View") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "FOV", FTString, length, 0) == -1)
        asfPrintError("Could not add FOV field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Resolution") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "RESOLUTION", FTString, length, 0) == -1)
        asfPrintError("Could not add RESOLUTION field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Overlap") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "OVERLAP", FTString, length, 0) == -1)
        asfPrintError("Could not add OVERLAP field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Scale") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "SCALE", FTString, length, 0) == -1)
        asfPrintError("Could not add SCALE field to database file\n");
    }
    else if (strcmp(dbf[ii].header, "Comments") == 0 && dbf[ii].visible) {
      if (DBFAddField(dbase, "COMMENTS", FTString, length, 0) == -1)
        asfPrintError("Could not add COMMENTS field to database file\n");
    }
  }

  // Close the database for initialization
  DBFClose(dbase);
  
  // Open shapefile for initialization
  shape = SHPCreate(inFile, SHPT_POLYGON);
  if (!shape)
    asfPrintError("Could not create shapefile '%s'\n", inFile);
  
  // Close shapefile for initialization
  SHPClose(shape);
  
  FREE(dbaseFile);

  return;
}

static void add_to_shape(DBFHandle dbase, SHPHandle shape, hap_t *hap,
			 dbf_header_t *dbf, int nCols, int n)
{
  int ii, field = 0;

  // Write fields into the database
  for (ii=0; ii<nCols; ii++) {
    if (strcmp(dbf[ii].header, "Filename") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->filename);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Date_Photography") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->date_photography);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Location") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->location);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Mission_Line") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->mission_line);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Roll") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->roll);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Frame") == 0 && dbf[ii].visible) {
      DBFWriteIntegerAttribute(dbase, n, field, hap->frame);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Center_Lat") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->center_lat);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Center_Lon") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->center_lon);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Upper_Left_Lat") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->near_start_lat);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Upper_Left_Lon") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->near_start_lon);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Upper_Right_Lat") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->far_start_lat);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Upper_Right_Lon") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->far_start_lon);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Lower_Left_Lat") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->near_end_lat);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Lower_Left_Lon") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->near_end_lon);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Lower_Right_Lat") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->far_end_lat);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Lower_Right_Lon") == 0 && dbf[ii].visible) {
      DBFWriteDoubleAttribute(dbase, n, field, hap->far_end_lon);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Photography") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->photography);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Camera") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->camera);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Focal_Length") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->focal_length);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Sensor_Coverage") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->sensor);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Altitude") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->altitude);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Field_Of_View") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->field_of_view);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Resolution") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->resolution);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Overlap") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->overlap);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Scale") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->scale);
      field++;
    }
    else if (strcmp(dbf[ii].header, "Comments") == 0 && dbf[ii].visible) {
      DBFWriteStringAttribute(dbase, n, field, hap->comments);
      field++;
    }
  }

  double lat[5], lon[5];
  lat[0] = lat[4] = hap->near_start_lat;
  lon[0] = lon[4] = hap->near_start_lon;
  lat[1] = hap->far_start_lat;
  lon[1] = hap->far_start_lon;
  lat[2] = hap->far_end_lat;
  lon[2] = hap->far_end_lon;
  lat[3] = hap->near_end_lat;
  lon[3] = hap->near_end_lon;  

  // Write shape object
  SHPObject *shapeObject=NULL;
  shapeObject = SHPCreateSimpleObject(SHPT_POLYGON, 5, lon, lat, NULL);
  SHPWriteObject(shape, -1, shapeObject);
  SHPDestroyObject(shapeObject);
}

int hap2shape(char *inFile, char *outFile, int listFlag)
{
  DBFHandle dbase;
  SHPHandle shape;
  hap_t hap;
  dbf_header_t *dbf;
  char *header, line[1024];
  int nCols, nColumns, ii=0;

  // Read configuration file
  read_header_config("HAP", &dbf, &nCols);

  // Read hap file
  FILE *ifp = FOPEN(inFile, "r");
  assert(ifp);
  check_hap_location(ifp, &header, &nColumns);
  
  // Initalize the database file
  shape_hap_init(outFile, header);
  open_shape(outFile, &dbase, &shape);

  if (listFlag) {
    while (fgets(line, 1022, ifp) != NULL) {
      strip_end_whitesp_inplace(line);
      
      // now get the individual column values
      hap_init(&hap);
      if (read_hap_line(header, nColumns, line, &hap)) {
	add_to_shape(dbase, shape, &hap, dbf, nCols, ii);
	ii++;
      }
      hap_free(&hap);
    }
  }
  else {
    fgets(line, 1022, ifp);
    strip_end_whitesp_inplace(line);
    hap_init(&hap);
    if (read_hap_line(header, nColumns, line, &hap))
      add_to_shape(dbase, shape, &hap, dbf, nCols, 0);
    hap_free(&hap);
  }

  // Clean up
  close_shape(dbase, shape);
  write_esri_proj_file(outFile);

  FCLOSE(ifp);

  return 1;
}
