#include "asf.h"
#include "shapefil.h"
#include "asf_vector.h"
#include <ctype.h>
#include <string.h>

static void remove_quotes(char *str)
{
  char *p = str;
  if (*p == '"')
    ++p;
  if (p[strlen(p)-1] == '"')
    p[strlen(p)-1] = '\0';
  strcpy(str, p);
}

// Initialize shape file
void shape_point_init(char *inFile)
{
  char *dbaseFile;
  DBFHandle dbase;
  SHPHandle shape;

  // Open database for initialization
  dbaseFile = (char *) MALLOC(sizeof(char)*(strlen(inFile)+5));
  sprintf(dbaseFile, "%s.dbf", inFile);
  dbase = DBFCreate(dbaseFile);
  if (!dbase)
    asfPrintError("Could not create database file '%s'\n", dbaseFile);

  // Add fields to database
  if (DBFAddField(dbase, "ID", FTString, 255, 0) == -1)
    asfPrintError("Could not add ID field to database file\n");
  if (DBFAddField(dbase, "LAT", FTDouble, 9, 4) == -1)
      asfPrintError("Could not add latitude field to database file\n");
  if (DBFAddField(dbase, "LON", FTDouble, 9, 4) == -1)
      asfPrintError("Could not add longitude field to database file\n");

  // Close the database for initialization
  DBFClose(dbase);

  // Open shapefile for initialization
  shape = SHPCreate(inFile, SHPT_POINT);
  if (!shape)
    asfPrintError("Could not create shapefile '%s'\n", inFile);

  // Close shapefile for initialization
  SHPClose(shape);
    
  FREE(dbaseFile);
  
  return;
}

// Convert point to kml file
int point2kml(char *inFile, char *outFile, int listFlag)
{
  float lat, lon;
  int points=0;
  char *p;
  char *line;
  char id[50];

  FILE *inFP = FOPEN(inFile, "r");
  FILE *outFP = FOPEN(outFile, "w");
  kml_header(outFP);  
  line = (char*)MALLOC(sizeof(char)*1024);
  points = 0;
  while (fgets(line, 1024, inFP)) {
    line[strlen(line)-1]='\0';
    p=line;
    while(isspace((int)(*p))) p++;
    if (*p != '#' && strlen(p) > 0) {
      // Extract information out of the line
      p = strchr(line, ',');
      if (p && *p == ',') {
	points++;
	sscanf(p+1, "%f,%f", &lat, &lon);
	*p = '\0';
	sprintf(id, "%s", line);
	remove_quotes(id);

	// Write information in kml file
	fprintf(outFP, "<Placemark>\n");
	fprintf(outFP, "  <name>%s</name>\n", line);
	fprintf(outFP, "  <description>\n");
	fprintf(outFP, "    <![CDATA[\n");
        fprintf(outFP, "<!-- Format: POINT (generated by %s) -->", version_string("convert2vector"));
	fprintf(outFP, "      <strong>ID</strong>: %s <br>\n", id);
	fprintf(outFP, "      <strong>Latitude</strong>: %9.4f <br>\n", lat);
	fprintf(outFP, "      <strong>Longitude</strong>: %9.4f <br>\n", lon);
	fprintf(outFP, "    ]]>\n");
	fprintf(outFP, "  </description>\n");
	fprintf(outFP, "  <LookAt>\n");
	fprintf(outFP, "    <longitude>%9.4f</longitude>\n", lon);
	fprintf(outFP, "    <latitude>%9.4f</latitude>\n", lat);
	fprintf(outFP, "    <range>400000</range>\n");
	fprintf(outFP, "  </LookAt>\n");
	fprintf(outFP, "  <Point>\n");
	fprintf(outFP, "    <coordinates>%f,%f,0</coordinates>\n", lon, lat);
	fprintf(outFP, "  </Point>\n");
	fprintf(outFP, "</Placemark>\n");
      }
    }
  }
  FREE(line);
  kml_footer(outFP);
  FCLOSE(inFP);
  FCLOSE(outFP);
  if (points < 1) {
    asfPrintError("Not enough points (lines) in point CSV file %s.\n", inFile);
  }

  return 1;
}

// Convert point to shapefile
int point2shape(char *inFile, char *outFile, int listFlag)
{
  DBFHandle dbase;
  SHPHandle shape;
  double lat, lon;
  char id[255], *p;
  int n;
  char *line;
  FILE *fp;

  // Initialize output
  shape_point_init(outFile);
  open_shape(outFile, &dbase, &shape);

  // Read point information out of file
  fp = FOPEN(inFile, "r");
  line = (char*)MALLOC(sizeof(char)*1024);
  n=0;
  while (fgets(line, 1024, fp)) {
    line[strlen(line)-1] = '\0';
    p = line;
    while (isspace((int)(*p))) p++;
    if (*p != '#') {
      p = strchr(line, ',');
      if (p && *p == ',') {
	int iid;
	sscanf(p+1, "%lf,%lf", &lat, &lon);
	*p = '\0';
	iid = strtol(line, (char**)NULL, 10);
	sprintf(id, "%d", iid);
	
	// Write information into database file
	DBFWriteStringAttribute(dbase, n, 0, id);
	DBFWriteDoubleAttribute(dbase, n, 1, lat);
	DBFWriteDoubleAttribute(dbase, n++, 2, lon);
	
	// Write shape object
	SHPObject *shapeObject=NULL;
	shapeObject = SHPCreateSimpleObject(SHPT_POINT, 1, &lon, &lat, NULL);
	SHPWriteObject(shape, -1, shapeObject);
	SHPDestroyObject(shapeObject);
      }
    }
  }
  FCLOSE(fp);

  // Clean up
  close_shape(dbase, shape);
  write_esri_proj_file(outFile);
  FREE(line);
  
  return 1;
}
