#include "asf.h"
#include "shapefil.h"
#include "asf_vector.h"
#include <ctype.h>
#include <string.h>

// Initialize shape file
void shape_polygon_init(char *inFile, int vertices)
{
  int ii;
  char *dbaseFile, header[10];
  DBFHandle dbase;
  SHPHandle shape;

  // Open database for initialization
  dbaseFile = (char *) MALLOC(sizeof(char)*(strlen(inFile)+5));
  sprintf(dbaseFile, "%s.dbf", inFile);
  dbase = DBFCreate(dbaseFile);
  if (!dbase)
    asfPrintError("Could not create database file '%s'\n", dbaseFile);

  // Add fields to database
  if (DBFAddField(dbase, "ID", FTString, 255, 0) == -1)
    asfPrintError("Could not add ID field to database file\n");
  if (vertices <= 10) {
    for (ii=0; ii<vertices; ii++) {
      sprintf(header, "LAT%d", ii+1);
      if (DBFAddField(dbase, header, FTDouble, 9, 4) == -1)
	asfPrintError("Could not add latitude field to database file\n");
      sprintf(header, "LON%d", ii+1);
      if (DBFAddField(dbase, header, FTDouble, 9, 4) == -1)
	asfPrintError("Could not add longitude field to database file\n");
    }
  }

  // Close the database for initialization
  DBFClose(dbase);

  // Open shapefile for initialization
  shape = SHPCreate(inFile, SHPT_POLYGON);
  if (!shape)
    asfPrintError("Could not create shapefile '%s'\n", inFile);

  // Close shapefile for initialization
  SHPClose(shape);

  FREE(dbaseFile);

  return;
}

// Convert point to kml file
int polygon2kml(char *inFile, char *outFile, int listFlag)
{
  int line_num, ii, vertices=0;
  float *lat = NULL, *lon = NULL; 
  char *p = NULL, *q = NULL;
  char *line = NULL;
  FILE *inFP = NULL, *outFP = NULL;
  
  inFP = FOPEN(inFile, "r");
  line = (char*)MALLOC(sizeof(char)*1024);

  // Count how many vertices are in the file
  vertices = 0;
  while (fgets(line, 1024, inFP)) {
    line[strlen(line)-1] = '\0';
    p = line;
    while(isspace((int)(*p))) p++;
    if (*p != '#') {
      p = strchr(line, ',');
      if (p && *p == ',') vertices++;
    }
  }
  if (vertices < 3) {
    FREE(line);
    FCLOSE(inFP);
    asfPrintError("Not enough polygon vertices (lines) in polygon CSV"
		  " file %s.\n", inFile);
  }
  
  // Read polygon vertices && ID line if it appears
  FSEEK(inFP, 0, SEEK_SET);
  lat = (float *) MALLOC(sizeof(float)*(vertices+1));
  lon = (float *) MALLOC(sizeof(float)*(vertices+1));
  ii = 0;
  line_num = 0;
  while (fgets(line, 1024, inFP)) {
    ++line_num;
    line[strlen(line)-1] = '\0';
    p = line;
    while (isspace(*p)) 
      p++;
    if (*p != '#' && *p != '\0') { // skip comment & blank lines
      int ncomma = count_char(p, ',');
      if (ncomma == 1) {
        // line has no ID field, just read in the lat/lon
        sscanf(p, "%f,%f", &lat[ii], &lon[ii]);
        ii++;
      }
      else if (ncomma == 2) {
        // line has an ID field -- skip past it
        q = strchr(p, ',');
        if (q) { // if should always pass
          sscanf(q+1, "%f,%f", &lat[ii], &lon[ii]);
          ii++;
        }
      }
      else {
        asfPrintStatus("Line %d: invalid -- ignored.\n", line_num);
      }
    }
  }
  FCLOSE(inFP);

  // in case the line contained some invalid lines, adjust nvertices
  vertices = ii;

  lat[vertices] = lat[0];
  lon[vertices] = lon[0];

  // Ball park center for <LookAt> position  - does not have to be accurate
  float clat=0.0, clon=0.0;
  for (ii=0; ii<vertices; ii++) {
    clat += lat[ii];
    clon += lon[ii];
  }
  clat /= vertices;
  clon /= vertices;
  
  // Write information in kml file
  outFP = FOPEN(outFile, "w");
  kml_header(outFP);
  fprintf(outFP, "<Placemark>\n");
  fprintf(outFP, "  <name>%s</name>\n", inFile);
  fprintf(outFP, "  <description>\n");
  fprintf(outFP, "    <![CDATA[\n");
  fprintf(outFP, "<!-- Format: POLYGON (generated by %s) -->\n",
          version_string("convert2vector"));
  fprintf(outFP, "      <strong>ID</strong>:%s <br>\n", inFile);
  fprintf(outFP, "      <strong>Vertices</strong>: %d<br>\n", vertices);
  for (ii=0; ii<vertices; ii++) {
    fprintf(outFP, "  <strong>Lat [%d]</strong>: %9.4f <br>\n", ii+1,lat[ii]);
    fprintf(outFP, "  <strong>Lon [%d]</strong>: %9.4f <br>\n", ii+1,lon[ii]);
  }
  fprintf(outFP, "    ]]>\n");
  fprintf(outFP, "  </description>\n");
  fprintf(outFP, "  <LookAt>\n");
  fprintf(outFP, "    <longitude>%9.4f</longitude>\n", clon);
  fprintf(outFP, "    <latitude>%9.4f</latitude>\n", clat);
  fprintf(outFP, "    <range>400000</range>\n");
  fprintf(outFP, "  </LookAt>\n");
  write_kml_style_keys(outFP);
  fprintf(outFP, "  <Polygon>\n");
  fprintf(outFP, "    <altitudeMode>relativeToGround</altitudeMode>\n");
  fprintf(outFP, "    <outerBoundaryIs>\n");
  fprintf(outFP, "      <LinearRing>\n");
  fprintf(outFP, "        <coordinates>\n");
  for (ii=0; ii<=vertices; ii++)
    fprintf(outFP, "         %.12f,%.12f,4000\n", lon[ii], lat[ii]);
  fprintf(outFP, "        </coordinates>\n");
  fprintf(outFP, "      </LinearRing>\n");
  fprintf(outFP, "    </outerBoundaryIs>\n");
  fprintf(outFP, "  </Polygon>\n");
  fprintf(outFP, "</Placemark>\n");
  kml_footer(outFP);

  FREE(lat);
  FREE(lon);
  FREE(line);
  FCLOSE(outFP);
  return 1;
}

// Write polygon to shapefile
static void polygon2shape_line(char *inFile, DBFHandle dbase, SHPHandle shape,
			       int n)
{
  FILE *fp;
  int vertices=0, foundFormat=FALSE;
  char poly[5], line[1024], *p;

  fp = FOPEN(inFile, "r");
  while (fgets(line, 1024, fp)) {
    if (line[0] != '#')
      vertices++;
    else {
      if (strstr(line, "Format"))
        p = strchr(line, ':');
      if (p) {
        p++;
        while (isspace(*p))
          p++;
        if (strncmp(p, "POLYGON", 7) == 0)
          foundFormat = TRUE;
      }
    }
  }
  FCLOSE(fp);  
  if (!foundFormat)
    asfPrintError("Input file is not a polygon CSV file.\n"
                  "Check 'convert2vector -help' for the correct format.\n");

  // Allocate memory
  double *lat = (double *) MALLOC(sizeof(double)*(vertices+1));
  double *lon = (double *) MALLOC(sizeof(double)*(vertices+1));

  // Write ID to shapefile
  sprintf(poly, "%d", n+1);
  DBFWriteStringAttribute(dbase, n, 0, poly);

  // Read file again: lat/lon
  int ii=0;
  fp = FOPEN(inFile, "r");
  while (fgets(line, 1024, fp)) {
    p = strchr(line, ',');
    if (p && line[0] != '#') {
      sscanf(p+1, "%lf,%lf", &lat[ii], &lon[ii]);
      // Write information into database file
      if (vertices <= 10) {
	DBFWriteDoubleAttribute(dbase, n, ii*2+1, lat[ii]);
	DBFWriteDoubleAttribute(dbase, n, ii*2+2, lon[ii]);
      }
      ii++;
    }
  }
  FCLOSE(fp);
  lat[vertices] = lat[0];
  lon[vertices] = lon[0];

  // Write shape object
  SHPObject *shapeObject=NULL;
  shapeObject =
    SHPCreateSimpleObject(SHPT_POLYGON, vertices+1, lon, lat, NULL);
  SHPWriteObject(shape, -1, shapeObject);
  SHPDestroyObject(shapeObject);

  // Clean up
  FREE(lat);
  FREE(lon);
}

// Convert polygon to shape
int polygon2shape(char *inFile, char *outFile, int listFlag)
{
  FILE *fp;
  DBFHandle dbase;
  SHPHandle shape;
  char line[1024], polyFile[1024], *p;
  int n=0, vertices=0, foundFormat=FALSE;

  // Determine the number of vertices
  // For lists take first file as a proxy
  if (listFlag) {
    fp = FOPEN(inFile, "r");
    while (fgets(line, 1024, fp)) {
      line[strlen(line)-1] = '\0';
      strip_end_whitesp_inplace(line);
      strcpy(polyFile, line);
    }
    FCLOSE(fp);
  }
  else
    strcpy(polyFile, inFile);
  fp = FOPEN(polyFile, "r");
  while (fgets(line, 1024, fp)) {
    if (line[0] != '#')
      vertices++;
    else {
      if (strstr(line, "Format"))
        p = strchr(line, ':');
      if (p) {
        p++;
        while (isspace(*p))
          p++;
        if (strncmp(p, "POLYGON", 7) == 0)
          foundFormat = TRUE;
      }
    }
  }
  FCLOSE(fp);  
  if (!foundFormat)
    asfPrintError("Input file is not a polygon CSV file.\n"
                  "Check 'convert2vector -help' for the correct format.\n");

  shape_polygon_init(outFile, vertices);
  open_shape(outFile, &dbase, &shape);

  if (listFlag) {
    fp = FOPEN(inFile, "r");
    while (fgets(line, 1024, fp)) {
      line[strlen(line)-1] = '\0';
      strip_end_whitesp_inplace(line);
      polygon2shape_line(line, dbase, shape, n);
      n++;
    }
    FCLOSE(fp);
  }
  else
    polygon2shape_line(inFile, dbase, shape, 0);

  close_shape(dbase, shape);
  write_esri_proj_file(outFile);

  return 1;
}
