/******************************************************************************
*                                             *
* Copyright (c) 2004, Geophysical Institute, University of Alaska Fairbanks   *
* All rights reserved.                                                        *
*                                                                             *
* Redistribution and use in source and binary forms, with or without          *
* modification, are permitted provided that the following conditions are met: *
*                                                                             *
*    * Redistributions of source code must retain the above copyright notice, *
*      this list of conditions and the following disclaimer.                  *
*    * Redistributions in binary form must reproduce the above copyright      *
*      notice, this list of conditions and the following disclaimer in the    *
*      documentation and/or other materials provided with the distribution.   *
*    * Neither the name of the Geophysical Institute nor the names of its     *
*      contributors may be used to endorse or promote products derived from   *
*      this software without specific prior written permission.               *
*                                                                             *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE    *
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
* POSSIBILITY OF SUCH DAMAGE.                                                 *
*                                                                             *
*       For more information contact us at:                                   *
*                                                                             *
*   Alaska Satellite Facility                                         *
*   Geophysical Institute           www.asf.alaska.edu            *
*       University of Alaska Fairbanks      uso@asf.alaska.edu        *
*   P.O. Box 757320                               *
*   Fairbanks, AK 99775-7320                          *
*                                         *
******************************************************************************/
/******************************************************************************
NAME:  sr2gr - converts slant range images to ground range images

SYNOPSIS:   sr2gr infile outmeta pixsiz

       infile     base name (.img,.meta)
       outfile    base name (.img,.meta)
       pixsiz     output pixel spacing (meters)

DESCRIPTION:
    This program converts slant range imagery into ground range imagery.
    The algorithm calculates the ground range to the first pixel in the
    image and from the spacecraft ephemeris, earth ellipsoid, and slant
    range to first pixel given in the image's metadata.  It then uses the
    slant range spacing interval to determine appropriate ground range
    positions.  The remapping is performed using bi-linear interpolation.

EXTERNAL ASSOCIATES:
    NAME:               USAGE:
    ---------------------------------------------------------------

FILE REFERENCES:
    NAME:               USAGE:
    ---------------------------------------------------------------

PROGRAM HISTORY:
    VERS:   DATE:  AUTHOR:      PURPOSE:
    ---------------------------------------------------------------
    1.0     6/97   T. Logan Original Implementation (Based on
                JPL provided routine sr2gr_vec)
    2.0     3/98   T. Logan     Modify to work with AISP outputs
    3.0     4/98   T. Logan     Modified to work with RAMMS data
    3.1     8/98   O. Lawlor    Corrected azimuth pixel spacing equation.
    4.0    12/98   O. Lawlor    Modified for new metadata routines.
    4.5     2/04   P. Denny     Update to use new metastruct instead of ddr
                                  Use improved get/put_float_line
    4.1     2/04   J. Nicoll    Initialize in buffer arrays to 0's because
                                  IRIX is too stupid to do it automatically.

HARDWARE/SOFTWARE LIMITATIONS:

ALGORITHM DESCRIPTION:

ALGORITHM REFERENCES:

BUGS:

******************************************************************************/

/****************************************************************************
*                                           *
*   sr2gr - converts slant range images to ground range images          *
* Copyright (c) 2004, Geophysical Institute, University of Alaska Fairbanks   *
* All rights reserved.                                                        *
*                                                                             *
* Redistribution and use in source and binary forms, with or without          *
* modification, are permitted provided that the following conditions are met: *
*                                                                             *
*    * Redistributions of source code must retain the above copyright notice, *
*      this list of conditions and the following disclaimer.                  *
*    * Redistributions in binary form must reproduce the above copyright      *
*      notice, this list of conditions and the following disclaimer in the    *
*      documentation and/or other materials provided with the distribution.   *
*    * Neither the name of the Geophysical Institute nor the names of its     *
*      contributors may be used to endorse or promote products derived from   *
*      this software without specific prior written permission.               *
*                                                                             *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE    *
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
* POSSIBILITY OF SUCH DAMAGE.                                                 *
*                                                                             *
*       For more information contact us at:                                   *
*                                                                             *
*       Alaska Satellite Facility                                             *
*       Geophysical Institute                   http://www.asf.alaska.edu     *
*       University of Alaska Fairbanks          uso@asf.alaska.edu            *
*       P.O. Box 757320                                                       *
*       Fairbanks, AK 99775-7320                                              *
*                                                                             *
******************************************************************************/

#include "asf_sar.h"
#include "asf.h"
#include "asf_license.h"
#include "sr2gr_help.h"

#include <stdio.h>
#include <stdlib.h>

#define VERSION 4.5
#define REQ_ARGS 3

int main(int argc,char *argv[])
{
    if (argc > 1) {
        check_for_help(argc, argv);
        handle_common_asf_args(&argc, &argv, TOOL_NAME);
    }
    if (argc != 4) {
      usage();
      return 1;
    }

    float grPixSize= -1;   /* output pixel size             */

    /* Make sure we've got the right amount of arguments */
    if (argc != 4) {usage(argv[0]);}

    asfPrintStatus("Converting from slant to ground range...\n");

    /* Get required arguments */
    grPixSize = atof(argv[3]);
    sr2gr_pixsiz(argv[1], argv[2], grPixSize);

    return 0;
}

